import time
from pathlib import Path

import torch

from ..utils.decorators import lazy
from ..callbacks.base import Callback
from ..viz import plot_line
from .. import resolve


class History(Callback):
  def __init__(self, *metrics, **named_metrics):
    super(History, self).__init__()
    self.metric_funcs = {
      **named_metrics,
    }
    for m in metrics:
      if isinstance(m, str):
        self.metric_funcs[m] = resolve.metric(m)
      elif hasattr(m, '__name__'):
        self.metric_funcs[m.__name__] = m
      else:
        raise ValueError(f'Unknown metric {m}')

    self.metrics = {m: [] for m in self.metric_funcs}
    self.metrics['loss'] = []
    self.times = []
    self.steps = []

    self.val_metrics = {m: [] for m in self.metric_funcs}
    self.val_metrics['loss'] = []
    self.val_steps = []
    self.val_times = []

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    self.times.append(time.time())
    self.steps.append(trainer.num_steps)
    self.metrics['loss'].append(loss.item())

    with torch.no_grad():
      for name, metric in self.metric_funcs.items():
        val = metric(targets, outputs).item()
        if torch.is_tensor(val):
          val = val.item()
        self.metrics[name].append(val)

  def on_validation_end(self, loss=None, outputs=None, targets=None,
                        trainer=None):
    self.val_times.append(time.time())
    self.val_steps.append(trainer.num_steps)
    self.val_metrics['loss'].append(loss.item())

    with torch.no_grad():
      for name, metric in self.metric_funcs.items():
        val = metric(targets, outputs).item()
        if torch.is_tensor(val):
          val = val.item()
        self.val_metrics[name].append(val)

  @lazy
  def plot(self):
    return HistoryPlotter(history=self)

  def summary(self):
    return {
      'train': {
        m: {
          'max': max(vals, default=None),
          'min': min(vals, default=None)
        }
        for m, vals in self.metrics.items()
      },
      'validation': {
        m: {
          'max': max(vals, default=None),
          'min': min(vals, default=None)
        }
        for m, vals in self.val_metrics.items()
      }
    }


class HistoryPlotter(Callback):
  def __init__(self, freq=500, window=50, metrics=None,
               clear=False, save=False, history: History = None):
    super().__init__()
    self.history: History = history
    self.freq = freq
    self.window = window
    self.metrics = metrics
    self.save = save

    self.figsize = (16, 16)

    self.clear = clear
    self.root = None

  def __call__(self, *args, **kwargs):
    return self.plot(*args, **kwargs)

  def plot(self, metric=None, time=False, validation=False, **kwargs):
    if self.clear:
      try:
        from IPython.display import clear_output
        clear_output(wait=True)
      except:
        pass

    if validation:
      ms, steps, times = (
        self.history.val_metrics,
        self.history.val_steps,
        self.history.val_times
      )
    else:
      ms, steps, times = (
        self.history.metrics,
        self.history.steps,
        self.history.times
      )

    if metric:
      metrics = [metric]
    else:
      metrics = self.metrics or ms.keys()

    for m in metrics:
      plot_line(
        ms[m],
        x=times if time else steps,
        xlabel='time' if time else 'step',
        ylabel=f'validation {m}' if validation else m,
        name=f'validation {m}' if validation else m,
        window=1 if validation else self.window,
        save=self.save and
             self.root / (f'validation {m}' if validation else m),
        show=not self.save,
        figsize=self.figsize,
        **kwargs
      )

  def on_train_start(self, trainer=None):
    if trainer:
      self.root = trainer.root
      self.history = self.history or trainer.history

  def on_batch_end(self, *args, trainer=None, **kwargs):
    if trainer.num_steps % self.freq == 0:
      self.plot(
        title=f'Epoch: {trainer.num_epochs} Steps: {trainer.num_steps}'
      )

  def on_validation_end(self, *args, trainer=None, **kwargs):
    self.plot(
      validation=True,
      title=f'Epoch: {trainer.num_epochs} Steps: {trainer.num_steps}'
    )


class HistoryWriter(Callback):
  def __init__(self, root=None, train=True, val=True, mode='a+',
               write_freq=1, flush_freq=500):
    self.root = root
    self.mode = mode
    self.train = train
    self.val = val

    self.flush_freq = flush_freq
    self.write_freq = write_freq

    self.header = None
    self.val_header = None

    self.train_file = None
    self.val_file = None

  @property
  def root(self):
    return self._root

  @root.setter
  def root(self, val):
    if val:
      self._root = Path(val)
      self._root.mkdir(parents=True, exist_ok=True)
    else:
      self._root = None

  def prep_files(self, root=None):
    self.root = self.root or root

    if self.train_file:
      self.train_file.close()
    if self.val_file:
      self.val_file.close()

    self.train_file = open(self.root / 'history-train.tsv', self.mode)
    self.val_file = open(self.root / 'history-val.tsv', self.mode)

  def on_train_start(self, trainer=None):
    self.prep_files(trainer.root if trainer else None)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if batch % self.write_freq:
      return

    if self.header is None:
      self.header = ['timestamp', 'epoch', 'step', *trainer.history.metrics]
      self.train_file.write('\t'.join(self.header) + '\n')

    self.train_file.write(
      f"{trainer.history.times[-1]}\t"
      f"{trainer.num_epochs}\t"
      f"{trainer.history.steps[-1]}\t"
      + '\t'.join((f"{trainer.history.metrics[m][-1]:.4f}"
                   for m in self.header[3:]))
      + '\n'
    )

    if batch % self.flush_freq == 0:
      self.train_file.flush()

  def on_validation_end(self, targets=None, outputs=None, loss=None,
                        trainer=None):
    if self.val_header is None:
      self.val_header = ['timestamp', 'epoch', 'step',
                         *trainer.history.val_metrics]
      self.val_file.write('\t'.join(self.header) + '\n')

    self.val_file.write(
      f"{trainer.history.val_times[-1]}\t"
      f"{trainer.num_epochs}\t"
      f"{trainer.history.val_steps[-1]}\t"
      + '\t'.join((f"{trainer.history.val_metrics[m][-1]:.4f}"
                   for m in self.val_header[3:]))
      + '\n'
    )

    self.val_file.flush()

  def on_train_end(self, trainer=None):
    self.train_file.close()
    self.val_file.close()

    self.train_file = None
    self.val_file = None

  def on_error(self, error, trainer=None):
    self.train_file.close()
    self.val_file.close()

    self.train_file = None
    self.val_file = None



import subprocess

def run(command):
  out = subprocess.check_output(command, shell=True)
  if out:
    return out.decode('utf-8')


def git_hash():
  return run('git rev-parse HEAD')


def git_commit(files='.', message='automated commit', branch=None):
  if branch:
    try:
      run(f'git checkout {branch}')
    except subprocess.SubprocessError:
      run(f'git checkout -b {branch}')
  if isinstance(files, str):
    run(f'git add {files}')
  else:
    run(['git', 'add', *files])
  run(f'git commit -m {message}')


def shutdown_computer():
  return run('sudo shutdown -h now')


def nvidia_smi():
  return run('nvidia-smi')


def pip_freeze():
  return run('pip freeze')


def conda_list(explicit=False):
  return run(f'conda list {"--explicit" if explicit else ""}')




from .shape import View, Flatten, Infer
from .stack import Stack


class Init:
  def __init__(self, cls, *args, **kwargs):
    self.cls = cls
    self.args = args
    self.kwargs = kwargs

  def __call__(self, *args, **kwargs):
    return self.cls(*self.args, *args, **{**self.kwargs, **kwargs})


import lmdb

from ..serialize import serialize, deserialize


class LMDB:
  NONE = '__LMDB_NONE__'

  def __init__(self, path, **kwargs):
    self.path = path
    self.db = None
    self.open(**kwargs)

  @property
  def stats(self):
    return self.db.stat()

  def __len__(self):
    return self.stats['entries']

  def close(self):
    self.db.close()

  def open(self, **kwargs):
    self.db = lmdb.open(self.path, **kwargs)

  def __del__(self):
    self.close()
    del self.db

  def __getitem__(self, key):
    with self.db.begin(write=False) as t:
      return self.deserialize(t.get(self.serialize(key)))

  def __setitem__(self, key, value):
    with self.db.begin(write=True) as t:
      return t.put(*self.serialize(key, value))

  def __delitem__(self, key):
    with self.db.begin(write=True) as t:
      return t.delete(self.serialize(key))

  def __iter__(self):
    with self.db.begin(write=False) as t:
      for k, v in t.cursor():
        yield self.deserialize(k, v)

  def serialize(self, key, value=NONE):
    if value is LMDB.NONE:
      return key
    else:
      return key, value

  def deserialize(self, key, value=NONE):
    if value is LMDB.NONE:
      return key
    else:
      return key, value



import torch
import numpy as np
from sklearn import metrics
from functools import partial
from collections import deque

from .utils import to_numpy


def get_preds(scores):
  score, preds = torch.max(scores, dim=1)
  return preds


def top_k(scores, k=5, largest=True):
  return torch.topk(scores, k=k, dim=1, largest=largest, sorted=True)


def accuracy(targets, preds):
  if targets.shape != preds.shape:
    preds = get_preds(preds)
  return (targets == preds).sum().float() / len(preds)


def top_k_accuracy(targets, preds, k=1):
  if len(targets.shape) != 1:
    raise ValueError('Multi label targets not supported')
  scores, preds = preds.topk(k, 1, True, True)
  preds = preds.t()
  correct = (preds == targets.view(1, -1).expand_as(preds))

  return correct.sum().float() / len(targets)


top_3_accuracy = partial(top_k_accuracy, k=3)
top_5_accuracy = partial(top_k_accuracy, k=5)
top_10_accuracy = partial(top_k_accuracy, k=10)


def precision_at_k(targets, outputs, k=5):
  scores, top_preds = top_k(outputs, k=k)
  raise NotImplementedError()


def recall_at_k(targets, outputs, k=5):
  scores, top_preds = top_k(outputs, k=k)
  raise NotImplementedError()


def hits_at_k():
  pass


def mean_reciprocal_rank():
  pass


def average_precision(targets, preds, target_threshold=0):
  targets, preds = to_numpy(targets), to_numpy(preds)
  if target_threshold is not None:
    targets = targets > target_threshold
  return metrics.average_precision_score(targets, preds)


def label_ranking_average_precision(targets, preds, target_threshold=0):
  targets, preds = to_numpy(targets), to_numpy(preds)
  if target_threshold is not None:
    targets = targets > target_threshold
  return metrics.label_ranking_average_precision_score(targets, preds)


def coverage_error(targets, preds, target_threshold=0):
  targets, preds = to_numpy(targets), to_numpy(preds)
  if target_threshold is not None:
    targets = targets > target_threshold
  return metrics.coverage_error(targets, preds)


def average_precision_at_k(targets, preds, k=5):
  raise NotImplementedError()


def mean_average_precision_at_k(targets, preds, k=5):
  raise NotImplementedError()


def evaluate_multiclass(
    targets,
    outputs,
    preds=None,
    classes=None
):
  preds = preds or get_preds(outputs)
  targets, outputs, preds = (
    to_numpy(targets),
    to_numpy(outputs),
    to_numpy(preds)
  )
  raise NotImplementedError()


def evaluate_multilabel(
    targets,
    outputs,
    preds=None,
    classes=None,
):
  targets, outputs, preds = (
    to_numpy(targets),
    to_numpy(outputs),
    to_numpy(preds)
  )
  raise NotImplementedError()


class Meter:
  def __init__(self):
    self.reset()

  def reset(self):
    self.max = None
    self.min = None
    self.sum = 0
    self.count = 0

  def update(self, val):
    if self.max is None or val > self.max:
      self.max = val
    if self.min is None or val < self.min:
      self.min = val

    self.sum += val
    self.count += 1

  def average(self):
    return self.sum / self.count


class MovingAvgMeter:
  def __init__(self, window=10):
    self.queue = deque(maxlen=window)

  def update(self, val):
    self.queue.append(val)

  def value(self):
    if not self.queue: return None
    return sum(self.queue) / len(self.queue)


def exp_moving_avg(cur, prev=None, alpha=.05, steps=None):
  """exponential moving average"""
  if prev is None:
    return cur
  avg = alpha * cur + prev * (1 - alpha)
  return avg / (1 - alpha ** steps) if steps else avg


def moving_average(data, window=10):
  cumsum = np.cumsum(np.insert(data, 0, 0))
  return (cumsum[window:] - cumsum[:-window]) / float(window)



from collections import OrderedDict
from collections import abc


class Container(abc.MutableMapping):
  def __init__(self, *args, **kwargs):
    items = OrderedDict(
      ('_arg' + str(n), v) for n, v in enumerate(args)
    )
    items.update(kwargs)

    self.__dict__.update(items)

    self._keys = list(items.keys())

    self.args = args
    self.kwargs = kwargs

  def __iter__(self):
    return (getattr(self, k) for k in self._keys)

  def __len__(self):
    return len(self._keys)

  def __getitem__(self, x):
    if isinstance(x, int):
      k = self._keys[x]
      return getattr(self, k)

    if isinstance(x, slice):
      return [getattr(self, k) for k in self._keys[x]]

  def __setitem__(self, key, value):
    if isinstance(key, int):
      setattr(self, self._keys[key], value)

  def __delitem__(self, key):
    pass


class Inputs(Container):
  pass


class Targets(Container):
  pass


class Outputs(Container):
  pass


class Samples:
  def __init__(self, inputs, targets):
    self.inputs = inputs
    self.targets = targets

  def __iter__(self):
    return (*self.inputs, *self.targets)



from yann.modules.conv.mixconv import MixConv
import torch

import pytest


def test_mixconv():
  c = MixConv(10, 20, (3,5,7))

  assert c.input_channel_counts == [4, 3, 3]
  assert c.output_channel_counts == [8, 6, 6]

  t = torch.rand((8, 10, 32, 32))

  x = c(t)

  assert x.shape == torch.Size([8, 20, 32, 32])

def test_not_depthwise():
  c = MixConv(10, 20, (3, 5), depthwise=False)
  t = torch.rand((8, 10, 32, 32))

  # NOTE: there's a bug in mkl-dnn that makes this fail
  #  when there's a filter of shape 7
  x = c(t)
  # assert False


def test_1_group():
  c = MixConv(4, 4, 3)

  assert len(c.convs) == 1

  t = torch.rand((10, 4, 32, 32))
  c(t)

def test_auto_groups():
  c = MixConv((2,2,2), (2,2,2))

  assert c.kernel_sizes == [3,5,7]

def test_variable_input_channel_counts():
  c = MixConv((16, 8, 4, 4), (16, 8, 4, 4), (3, 5, 7, 9))

  t = torch.rand((8, 32, 32, 32))

  x = c(t)

  assert x.shape

  c = MixConv((16, 8, 4, 4), (8, 4, 2, 2), (3, 5, 7, 9))

  c(t)

  c = MixConv((16, 8, 4, 4), (16, 16, 4, 4), (3, 5, 7, 9))
  c(t)

  c = MixConv((16, 8, 4, 4), (16, 16, 4, 4), (3, 5, 7))
  c(t)


import torch
import gc
import numpy as np
from torch import nn


def iter_allocated_tensors():
  for x in gc.get_objects():
    try:
      if torch.is_tensor(x) or (hasattr(x, 'data') and torch.is_tensor(x.data)):
        yield x
    except:
      pass


def bytecount(tensor: torch.Tensor):
  return tensor.numel() * tensor.element_size()


def receptive_field(module: nn.Module, input_shape=(1, 1, 256, 256)):
  module.zero_grad()
  input = torch.ones(*input_shape, requires_grad=True)
  output = module(input)
  indicator = torch.zeros_like(output)
  indicator[(0, 0, *(n // 2 for n in indicator.shape[2:]))] = 1

  loss = torch.mean(output * indicator)
  loss.backward()

  grad = input.grad.numpy()
  nonzero_indices = np.where(grad != 0)
  return tuple(x.max() - x.min() + 1 for x in nonzero_indices)


from typing import Dict



class Blueprint:
  @classmethod
  def from_state_dict(cls, d):
    pass

  def state_dict(self):
    pass

  def save(self):
    pass

  def load(self):
    pass

  def info(self) -> Dict:
    pass

  def train(self):
    pass

  def predict(self, inputs):
    pass

  def evaluate(self):
    pass


import os


class ConfigError(Exception):
  pass


class Vars:
  def __init__(self, env_prefix=None):
    self.env_prefix = env_prefix

  def __call__(self, option, required=False, cast=None, null=False):
    env_var_name = '_'.join((self.env_prefix, option)) \
      if self.env_prefix else option
    if env_var_name in os.environ:
      val = os.environ[env_var_name]
      val = cast(val) if cast else val
      if not null and not val:
        raise ConfigError(
          f"{option} cannot be null"
        )
      return val

    if hasattr(self, option):
      val = getattr(self, option)
      if not null and not val:
        raise ConfigError(
          f"{option} cannot be null"
        )
      return val

    if required:
      raise ConfigError(
        f"Required configuration value was not provided ({env_var_name})"
      )

  def __contains__(self, item):
    return item in os.environ or hasattr(self, item)


class YannConfig(Vars):
  BATCH_SIZE = 34
